import json
import os
from typing import Optional
from urllib.parse import urlparse

import click
import requests
import requests.auth
from click_aliases import ClickAliasedGroup
from requests.exceptions import RequestException
from requests_toolbelt.sessions import BaseUrlSession

from . import auth, config, index, mc, node, root, s3token, search, user, workspace


class WioSession(BaseUrlSession):
    def __init__(self, cfg: config.Config):
        base_url = cfg.api_url
        base_url = (
            f'{base_url.rstrip("/")}/'  # tolerate input with or without trailing slash
        )
        super(WioSession, self).__init__(base_url=base_url)
        auth = None
        if cfg.access_key and cfg.secret_key:
            auth = (cfg.access_key, cfg.secret_key)
            self.headers.update(
                {
                    "Authorization": requests.auth._basic_auth_str(
                        cfg.access_key, cfg.secret_key
                    ),
                }
            )
        self.headers.update(
            {
                "User-agent": f"wio",
                "Accept": "application/json",
            }
        )


@click.group(cls=ClickAliasedGroup)
@click.option("--api-url", envvar="WIO_ENDPOINT_URL")
@click.option(
    "--config-path",
    default=os.path.join(os.path.expanduser("~"), ".config", "wio.json"),
    envvar="WIO_CONFIG_PATH",
    type=click.Path(dir_okay=False, file_okay=True, writable=True, resolve_path=True),
)
@click.version_option()
@click.pass_context
def cli(ctx, api_url, config_path):
    conf = config.load_config(config_path)
    if api_url:
        conf.api_url = api_url
    ctx.obj = {
        "configPath": config_path,
        "config": conf,
        "session": WioSession(conf),
    }


workspace.make(cli)
auth.make(cli)
s3token.make(cli)
mc.make(cli)
index.make(cli)
node.make(cli)
root.make(cli)
search.make(cli)
user.make(cli)



import webbrowser
from typing import cast

import click

from workspacesio.common import schemas

from . import config
from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @click.command(name="login")
    @click.option("--access-key", prompt=True)
    @click.option("--secret-key", prompt=True)
    @click.pass_obj
    def login(ctx, access_key, secret_key):
        ctx = config.getctx(ctx)
        r = ctx.session.get("user/me", auth=(access_key, secret_key))
        if r.ok:
            click.echo(click.style("Login success", fg="green", bold=True))
            ctx.config.access_key = access_key
            ctx.config.secret_key = secret_key
            config.save(ctx)
        else:
            exit_with(handle_request_error(r))

    @click.command(name="info")
    @click.pass_obj
    def me(ctx):
        ctx = config.getctx(ctx)
        r1 = ctx.session.get("user/me")
        if not r1.ok:
            exit_with(handle_request_error(r1))
        out = r1.json()
        out.update(ctx.config.dict())
        out.update({"configPath": ctx.configPath})
        exit_with(out)

    cli.add_command(login)
    cli.add_command(me)



import os
import re
import sys
import urllib.parse
from typing import Dict, List, Tuple

import click

from workspacesio.common import schemas

from .util import exit_with, handle_request_error

SUPPORTED_MINIO_COMMANDS = [
    "ls",
    "cp",
    "mirror",
    "cat",
    "head",
    "pipe",
    "share",
    "find",
    "sql",
    "stat",
    "mv",
    "tree",
    "du",
    "diff",
    "rm",
    "watch",
]


def isrealpath(s: str) -> bool:
    return os.path.exists(os.path.abspath(os.path.expanduser(s)))


def transform_arguments(args: List[str]):
    """attempt to """
    if len(args) < 2 or not (args[0] in SUPPORTED_MINIO_COMMANDS):
        return []
    possible_workspaces = []
    for arg in args[1:]:
        if arg.startswith("-"):
            continue
        if isrealpath(arg):
            continue
        possible_workspaces.append(arg)
    return possible_workspaces


def make(cli: click.Group):
    @cli.command(
        name="mc",
        context_settings=dict(
            ignore_unknown_options=True,
        ),
    )
    @click.argument("args", nargs=-1)
    @click.pass_obj
    def mc(ctx, args):
        from workspacesio.common import s3utils

        r = ctx["session"].post(
            "token/search",
            json={
                "search_terms": args,
            },
        )
        if r.ok:
            response = r.json()
            assembled = " ".join(args)
            mc_env = ""
            for arg, match in response["workspaces"].items():
                workspace = schemas.WorkspaceDB(**match["workspace"])
                scope = workspace.root.root_type.lower()
                key = s3utils.getWorkspaceKey(workspace)
                path = "/".join(
                    [
                        "myalias",
                        workspace.root.bucket,
                        key,
                        match["path"].lstrip("/"),
                    ]
                )
                assembled = assembled.replace(arg, path)
            if len(response["tokens"]) == 1:
                token = response["tokens"][0]["token"]
                access_key = token["access_key_id"]
                secret = token["secret_access_key"]
                session_token = token["session_token"]
                api_url = response["tokens"][0]["node"]["api_url"]
                url = urllib.parse.urlparse(api_url)
                mc_env = (
                    f"{url.scheme}://{access_key}:{secret}:{session_token}@{url.netloc}"
                )
            command = (
                "mc",
                *assembled.split(" "),
            )
            os.execvpe(command[0], command, dict(os.environ, MC_HOST_myalias=mc_env))
        else:
            exit_with(handle_request_error(r))



"""
Produce sources for indexing given some root
"""
import posixpath
from typing import Generator, Iterable, List, Optional, Tuple, TypeVar, Union

import minio

from . import indexing_schemas, s3utils, schemas, video

clientCache = s3utils.Boto3ClientCache()


def minio_list_root_children(
    node: schemas.StorageNodeOperator, root: schemas.WorkspaceRootDB
) -> dict:
    """Get direct children from a root"""
    prefix = posixpath.join(root.base_path, "")  # add trailing slash
    return clientCache.get_minio_sdk_client(node).list_objects_v2(
        root.bucket, prefix=prefix
    )


def minio_recursive_generate_objects(
    node: schemas.StorageNodeOperator,
    root: schemas.WorkspaceRootDB,
    workspace: schemas.WorkspaceDB,
    after: str = "",
) -> Iterable[minio.Object]:
    """Generate a flat list of minio objects from a workspace"""
    b3client = clientCache.get_minio_sdk_client(node)
    bucket = root.bucket
    prefix = posixpath.join(root.base_path, s3utils.getWorkspaceKey(workspace))
    start_after = posixpath.join(prefix, after.lstrip("/"))
    return b3client.list_objects_v2(
        bucket,
        prefix=prefix,
        recursive=True,
        start_after=start_after,
    )


def minio_buffer_objects(
    objects: List[minio.Object], buffer_size=10
) -> Iterable[minio.Object]:
    """Return small batches of objects"""
    index = 0
    buf: List[minio.Object] = []
    for o in objects:
        if index < buffer_size:
            buf.append(o)
            index += 1
        else:
            yield buf
            index = 0
            buf = []
    if index != 0:
        yield buf


def minio_transform_object(
    workspace: schemas.WorkspaceDB, root: schemas.WorkspaceRootDB, obj: minio.Object
) -> indexing_schemas.IndexDocumentBase:
    """Turn an object into a index document"""
    common = s3utils.getWorkspaceKey(workspace)
    assert (
        posixpath.commonprefix([common, obj.object_name]) is common
    ), f"{common} not in {obj.object_name}"
    inner = obj.object_name.lstrip(common)
    return indexing_schemas.IndexDocumentBase(
        time=obj.last_modified,
        size=obj.size,
        etag=obj.etag,
        path=inner,
        extension=posixpath.splitext(inner)[-1],
        filename=posixpath.basename(obj.object_name),
        content_type=obj.content_type,
    )


def additional_indexes(
    node: schemas.StorageNodeOperator,
    root: schemas.WorkspaceRootDB,
    workspace: schemas.WorkspaceDB,
    doc: indexing_schemas.IndexDocumentBase,
) -> Tuple[List[str], List[str]]:
    """Produce additional indexes on the document if it is supported"""
    analyses_succeeded = []
    analyses_failed = []
    if doc.extension in [".mp4", ".avi", ".mkv", ".webm", ".wmv"]:
        try:
            video.probe(doc=doc, node=node, root=root, workspace=workspace)
            analyses_succeeded.append("ffprobe")
        except RuntimeError:
            analyses_failed.append("ffprobe")
    # if doc.extension in ['.csv', '.txt', '.yml', '']
    return analyses_succeeded, analyses_failed



import click

from . import config
from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.command(name="search")
    @click.argument("query", type=click.STRING)
    @click.pass_obj
    def search(ctx, query):
        ctx = config.getctx(ctx)
        r = ctx.session.get(
            "search",
            params={"q": query},
        )
        if not r.ok:
            exit_with(handle_request_error(r))
        for result in r.json()["hits"]["hits"]:
            click.secho(
                f'{result["_source"]["owner_name"]}/', fg="cyan", bold=True, nl=False
            )
            click.secho(
                f'{result["_source"]["workspace_name"]}',
                fg="cyan",
                bold=True,
                nl=False,
            )
            click.secho(result["_source"]["path"])

    cli.add_command(search)



import click
from click_aliases import ClickAliasedGroup

from workspacesio.common import schemas

from . import config
from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.group(name="user", cls=ClickAliasedGroup, aliases=["u"])
    def user():
        pass

    @user.command(name="list", aliases=["l", "ls"])
    @click.pass_obj
    def findusers(ctx):
        ctx = config.getctx(ctx)
        r = ctx.session.get(f"user")
        if r.ok:
            for user in r.json():
                userschema = schemas.UserDB(**user)
                click.secho(f"[{userschema.created}] ", fg="green", nl=False)
                click.secho(f"{userschema.id} ", fg="yellow", nl=False)
                click.secho(f"{userschema.username} ", fg="cyan", bold=True)
        else:
            exit_with(handle_request_error(r))



from datetime import datetime

import click
from click_aliases import ClickAliasedGroup

from workspacesio.common import schemas

from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.group(name="token", cls=ClickAliasedGroup, aliases=["t"])
    def token():
        pass

    @token.command(name="fetch", aliases=["f"])
    @click.argument("workspaces", type=click.STRING, nargs=-1)
    @click.pass_obj
    def create_token(ctx, workspaces):
        if len(workspaces) == 0:
            return
        r = ctx["session"].post("token/search", json={"search_terms": workspaces})
        if r.ok:
            response = schemas.S3TokenSearchResponse(**r.json())
            for wrapper in response.tokens:
                node = wrapper.node
                token = wrapper.token
                click.secho(
                    f"Credentials for {[w.name for w in token.workspaces]} @ {node.api_url}",
                    fg="green",
                )
                click.secho(
                    f"Expires in {token.expiration - datetime.utcnow()}\n", fg="yellow"
                )
                click.secho(f"export AWS_ACCESS_KEY_ID={token.access_key_id}")
                click.secho(f"export AWS_SECRET_ACCESS_KEY={token.secret_access_key}")
                click.secho(f"export AWS_SESSION_TOKEN={token.session_token}\n")
        else:
            exit_with(handle_request_error(r))

    @token.command(name="list", aliases=["l", "ls"])
    @click.pass_obj
    def list_token(ctx):
        r = ctx["session"].get("token")
        exit_with(handle_request_error(r))

    @token.command(name="delete", aliases=["d"])
    @click.option("--all", is_flag=True)
    @click.argument("token_id", required=False)
    @click.pass_obj
    def delete_token(ctx, all, token_id):
        if all:
            r = ctx["session"].delete("token")
        else:
            r = ctx["session"].delete(f"token/{token_id}")
        exit_with(handle_request_error(r))

    cli.add_command(token)



import click
from click_aliases import ClickAliasedGroup

from workspacesio.common import schemas

from . import config
from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.group(name="root", cls=ClickAliasedGroup, aliases=["r"])
    def root():
        pass

    @root.command(name="list", aliases=["l", "ls"])
    @click.option("--node-name", type=click.STRING)
    @click.pass_obj
    def list_roots(ctx, node_name):
        r = ctx["session"].get("root", params={"node_name": node_name})
        exit_with(handle_request_error(r))

    @root.command(name="create", aliases=["c"])
    @click.argument("bucket", type=click.STRING)
    @click.argument("node_name", type=click.STRING)
    @click.option("--base-path", type=click.STRING, default="")
    @click.option(
        "--root-type",
        type=click.Choice(schemas.RootType),
        default=schemas.RootType.PRIVATE.value,
    )
    @click.pass_obj
    def create_root(ctx, bucket, node_name, base_path, root_type):
        r = ctx["session"].post(
            "root",
            json={
                "bucket": bucket,
                "node_name": node_name,
                "base_path": base_path,
                "root_type": root_type,
            },
        )
        exit_with(handle_request_error(r))

    @root.command(name="delete", aliases=["d"])
    @click.argument("root_id", type=click.STRING)
    @click.pass_obj
    def delete_root(ctx, root_id):
        r = ctx["session"].delete(f"root/{root_id}")
        exit_with(handle_request_error(r))

    @root.command(name="import", help="Import all workspaces in a root.")
    @click.argument("root_id")
    @click.option("--index-all", is_flag=True)
    @click.pass_obj
    def import_all_workspaces(ctx, root_id, index_all):
        # Dynamic, expensive imports
        from workspacesio.common import producers

        ctx = config.getctx(ctx)
        r = ctx.session.post(f"root/{root_id}/import")
        if not r.ok:
            exit_with(handle_request_error(r))
        rdata = schemas.RootCredentials(**r.json())
        root_contents = producers.minio_list_root_children(
            node=rdata.node, root=rdata.root
        )
        workspace_list: List[schemas.WorkspaceDB] = []
        for folder in root_contents:
            prefix = folder.object_name.lstrip(rdata.root.base_path).strip("/")
            if len(prefix) > 0:
                print(f"Discovered {prefix}")
                workspace = ctx.session.post(
                    "workspace",
                    json={
                        "name": prefix,
                        "public": False,
                        "unmanaged": True,
                        "base_path": prefix,
                        "node_name": rdata.node.name,
                        "root_id": str(rdata.root.id),
                    },
                )
                if not workspace.ok and workspace.status_code != 409:
                    exit_with(handle_request_error(workspace))
                workspace_list.append(schemas.WorkspaceDB(**workspace.json()))

    @root.command(
        name="import-workspace", help="Import a particular prefix as a workspace."
    )
    @click.argument("root_id", type=click.STRING)
    @click.argument("--base-path", type=click.STRING, default="")
    def import_workspace(ctx, root_id, base_path):
        pass

    cli.add_command(root)



import click
from click_aliases import ClickAliasedGroup

from workspacesio.common import schemas

from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.group(name="node", cls=ClickAliasedGroup, aliases=["n"])
    def node():
        pass

    @node.command(name="list", aliases=["l", "ls"])
    @click.pass_obj
    def ls(ctx):
        r = ctx["session"].get("node")
        if r.ok:
            for node in r.json():
                api_url = node["api_url"]
                click.secho(f"[{node['created']}] ", fg="green", nl=False)
                click.secho(f"{node['id']} ", fg="yellow", nl=False)
                click.secho(f"{api_url} ", fg="bright_black", nl=False)
                click.secho(f"{node['name']}", fg="cyan", bold=True)
        else:
            exit_with(handle_request_error(r))

    @node.command(name="delete", aliases=["d"])
    @click.argument("node_id", type=click.STRING, required=True)
    @click.pass_obj
    def rm(ctx, node_id):
        r = ctx["session"].delete(f"node/{node_id}")
        exit_with(handle_request_error(r))

    @node.command(name="create", aliases=["c"])
    @click.argument("name", type=click.STRING)
    @click.argument("api_url", type=click.STRING)
    @click.argument("access_key_id", type=click.STRING)
    @click.argument("secret_access_key", type=click.STRING)
    @click.option("--region-name", type=click.STRING, default="us-east-1")
    @click.option("--sts-api-url", type=click.STRING)
    @click.option(
        "--role-arn",
        type=click.STRING,
        help="ARN for role to use during STS assume role.  Required for s3, Ignored for MinIO.  Make sure this role has NO permissions",
    )
    @click.pass_obj
    def register(
        ctx,
        name,
        api_url,
        access_key_id,
        secret_access_key,
        region_name,
        sts_api_url,
        role_arn,
    ):
        r = ctx["session"].post(
            "node",
            json={
                "name": name,
                "api_url": api_url,
                "access_key_id": access_key_id,
                "secret_access_key": secret_access_key,
                "region_name": region_name,
                "sts_api_url": sts_api_url,
                "assume_role_arn": role_arn,
            },
        )
        exit_with(handle_request_error(r))

    cli.add_command(node)



import posixpath

import click

from . import config
from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.group(name="index")
    def index():
        pass

    @index.command(name="delete")
    @click.argument("root_id", type=click.STRING)
    @click.pass_obj
    def delete(ctx, root_id):
        r = config.getctx(ctx).session.delete(f"root/{root_id}/index")
        exit_with(handle_request_error(r))

    @index.command(name="create")
    @click.argument("root_id", type=click.STRING)
    @click.pass_obj
    def create(ctx, root_id):
        ctx = config.getctx(ctx)
        r = ctx.session.post(f"root/{root_id}/index")
        r2 = ctx.session.get("info")
        if r.ok and r2.ok:
            data = r.json()
            infodata = r2.json()
            root_id = data["root_id"]
            bucket = data["root"]["bucket"]
            prefix = posixpath.join(data["root"]["base_path"], "/").lstrip("/")
            endpoint_base = infodata["public_address"]
            click.secho(
                "To notify Workspaces of updates, configure your MinIO instance using these commands.\n",
                fg="yellow",
            )
            for c in [
                # wehook ID should come from ROOT, not index.  You only want to subscribe to events once.
                f"export ALIAS=local",
                f"mc admin config set $ALIAS notify_webhook:{root_id} endpoint={endpoint_base}/api/minio/events enable=on",
                f"mc event add $ALIAS/{bucket} arn:minio:sqs::{root_id}:webhook {f'--prefix {prefix}' if prefix else ''} --event delete,put",
            ]:
                click.secho("\t" + c, fg="blue", bold=True)
            click.echo("")
        else:
            exit_with(handle_request_error(r))

    cli.add_command(index)



import json
from collections.abc import Iterable
from json.decoder import JSONDecodeError
from typing import List

import click
from requests import Response


def handle_request_error(r: Response) -> dict:
    if not r.ok:
        # TODO: inspect content-type to infer this
        error_text = r.text
        try:
            error_text = r.json()
        except ValueError:
            pass

        headers = dict(r.request.headers)
        headers.pop("Authorization", None)
        body = str(r.request.body)
        try:
            body = json.loads(body)
        except JSONDecodeError:
            pass

        if r.status_code == 401:
            return {"error": "You are not logged in."}

        return {
            "context": {
                "url": r.url,
                "method": r.request.method,
                "status": r.status_code,
                "body": body,
                "headers": headers,
            },
            "error": error_text,
        }
    return {"response": r.json()}


def exit_with(out: dict):
    if out.get("error"):
        click.secho(json.dumps(out, indent=2, sort_keys=True), fg="red")
        exit(1)
    click.echo(json.dumps(out, indent=2, sort_keys=True))
    exit(0)



import datetime
import enum
import uuid
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel


class DBBaseModel(BaseModel):
    id: uuid.UUID
    created: datetime.datetime

    class Config:
        orm_mode = True


class ShareType(str, enum.Enum):
    """
    READ holders can read
    WRITE holders can read and write
    OWN holders can read, write, delete, and share
    """

    READ = "read"
    READWRITE = "readwrite"
    OWN = "own"


class RootType(str, enum.Enum):
    PUBLIC = "public"
    PRIVATE = "private"
    UNMANAGED = "unmanaged"


class ServerInfo(BaseModel):
    public_address: str


###########################################################
# User Schemas
###########################################################


class UserBase(BaseModel):
    username: str


class UserDB(DBBaseModel, UserBase):
    pass


###########################################################
# Storage Node Schemas
###########################################################


class StorageNodeBase(BaseModel):
    name: str
    api_url: str
    sts_api_url: Optional[str]
    region_name: str


class StorageNodeCreate(StorageNodeBase):
    access_key_id: str
    secret_access_key: str
    assume_role_arn: Optional[str]


class StorageNodeDB(DBBaseModel, StorageNodeBase):
    creator_id: uuid.UUID
    creator: UserDB


class StorageNodeOperator(StorageNodeDB):
    access_key_id: str
    secret_access_key: str
    assume_role_arn: Optional[str]


###########################################################
# Workspace Root Schemas
###########################################################


class WorkspaceRootBase(BaseModel):
    root_type: RootType
    bucket: str
    base_path: str


class WorkspaceRootCreate(WorkspaceRootBase):
    node_name: str


class WorkspaceRootDB(DBBaseModel, WorkspaceRootBase):
    node_id: uuid.UUID


class WorkspaceRootOperator(WorkspaceRootDB):
    storage_node: StorageNodeDB


###########################################################
# RootCredentials Schemas
###########################################################


class RootCredentials(BaseModel):
    """
    Only passed out to node operators
    and administrators
    """

    root: WorkspaceRootOperator
    node: StorageNodeOperator


###########################################################
# Workspace Schemas
###########################################################


class WorkspaceBase(BaseModel):
    name: str
    base_path: Optional[str]


class WorkspaceCreate(WorkspaceBase):
    public: Optional[bool] = False
    unmanaged: Optional[bool] = False
    node_name: Optional[str]
    root_id: Optional[uuid.UUID]


class WorkspaceDB(DBBaseModel, WorkspaceBase):
    id: uuid.UUID
    name: str
    owner_id: uuid.UUID
    root_id: uuid.UUID
    owner: UserDB
    root: WorkspaceRootDB


###########################################################
# API Key Schemas
###########################################################


class ApiKeyBase(BaseModel):
    pass


class ApiKeyCreate(ApiKeyBase):
    pass


class ApiKeyCreateResponse(DBBaseModel, ApiKeyBase):
    key_id: str
    secret: str
    user_id: uuid.UUID


class ApiKeyDB(DBBaseModel, ApiKeyBase):
    key_id: str
    user_id: uuid.UUID

    user: UserDB


###########################################################
# S3 Token Schemas
###########################################################


class S3TokenBase(BaseModel):
    expiration: Optional[datetime.datetime]


class S3TokenCreate(S3TokenBase):
    workspaces: List[uuid.UUID]


class S3TokenSearch(BaseModel):
    search_terms: List[str]


class S3TokenDB(DBBaseModel, S3TokenBase):
    owner_id: uuid.UUID
    storage_node_id: uuid.UUID
    access_key_id: str
    secret_access_key: str
    session_token: str
    policy: dict
    owner: UserDB
    workspaces: List[WorkspaceDB]


class S3TokenSearchResponseWorkspacePart(BaseModel):
    path: str
    workspace: WorkspaceDB


class TokenNodeWrapper(BaseModel):
    token: S3TokenDB
    node: StorageNodeDB


class S3TokenSearchResponse(BaseModel):
    tokens: List[TokenNodeWrapper]
    workspaces: Dict[str, S3TokenSearchResponseWorkspacePart]


###########################################################
# Share Schemas
###########################################################


class ShareBase(BaseModel):
    permission: ShareType
    expiration: Optional[datetime.datetime]


class ShareCreate(ShareBase):
    sharee_id: Optional[uuid.UUID]
    workspace_id: Optional[uuid.UUID]
    sharee: Optional[str]
    workspace: Optional[str]


class ShareUpdate(ShareBase):
    pass


class ShareDB(DBBaseModel, ShareBase):
    creator_id: uuid.UUID
    workspace_id: uuid.UUID
    sharee_id: uuid.UUID
    workspace: WorkspaceDB
    creator: UserDB
    sharee: UserDB


###########################################################
# Instrumentation Schemas
###########################################################


class InstrumentBase(BaseModel):
    tag: str
    enabled: bool


class InstrumentUpdate(InstrumentBase):
    pass


class InstrumentDB(InstrumentBase, DBBaseModel):
    last_updated: datetime.datetime
    pulled_size: int
    pushed_size: int
    pulled_count: int
    pushed_count: int



import os

from setuptools import find_packages, setup

deps = [
    "boto3",
    "click",
    "click-aliases",
    "colorama",
    "elasticsearch>=7.0.0,<8.0.0",
    "ffmpeg-python",
    "minio",
    "pydantic",
    "requests",
    "requests-toolbelt",
    "tqdm",
]

setup(
    name="workspacesio-cli",
    version="0.1.0",
    script_name="setuptools.wio.py",
    python_requires=">3.7",
    zip_safe=False,
    install_requires=deps,
    include_package_data=True,
    packages=["workspacesio.cli", "workspacesio.common"],
    entry_points={
        "console_scripts": [
            "wio=workspacesio.cli:cli",
        ],
    },
)



import datetime
import json
from typing import List

import click
from click_aliases import ClickAliasedGroup
from tqdm import tqdm

from workspacesio.common import indexing_schemas, schemas

from . import config
from .util import exit_with, handle_request_error


def make(cli: click.Group):
    @cli.group(name="workspace", cls=ClickAliasedGroup, aliases=["w"])
    def workspace():
        pass

    @workspace.command(name="list", aliases=["ls", "l"])
    @click.option("--name", type=click.STRING, required=False)
    @click.option("--like", type=click.STRING, required=False)
    @click.option("--public", is_flag=True)
    @click.pass_obj
    def list_workspaces(ctx, name, like, public):
        params = {"public": public}
        if name:
            params["name"] = name
        if like:
            params["like"] = like
        r = ctx["session"].get("workspace", params=params)
        if r.ok:
            for ws in r.json():
                scope = ws["root"]["root_type"]
                click.secho(f"[{ws['created']}] ", fg="green", nl=False)
                click.secho(f"{ws['id']} ", fg="yellow", nl=False)
                click.secho(
                    f"{ws['owner']['username']}/{ws['name']}/ ",
                    fg="cyan",
                    bold=True,
                    nl=False,
                )
                click.secho(f"({scope})", fg="bright_black")
        else:
            exit_with(handle_request_error(r))

    @workspace.command(name="create", aliases=["c"])
    @click.argument("name")
    @click.option("--public/--private", default=False, is_flag=True)
    @click.option("--unmanaged", default=False, is_flag=True)
    @click.option("--node-name", type=click.STRING, default=None)
    @click.pass_obj
    def create_workspace(ctx, name, public, unmanaged, node_name):
        r = ctx["session"].post(
            "workspace",
            json={
                "name": name,
                "public": public,
                "unmanaged": unmanaged,
                "node_name": node_name,
            },
        )
        exit_with(handle_request_error(r))

    @workspace.command(name="delete")
    @click.argument("workspace_id", type=click.STRING)
    @click.pass_obj
    def delete_workspace(ctx, workspace_id):
        r = ctx["session"].delete(f"workspace/{workspace_id}")
        exit_with(handle_request_error(r))

    @workspace.command(name="share", aliases=["s"])
    @click.argument("workspace", type=click.STRING)
    @click.argument("sharee", type=click.STRING)
    @click.option(
        "--permission",
        type=click.Choice(schemas.ShareType),
        default=schemas.ShareType.READ.value,
    )
    @click.option("--expire", type=click.DateTime())
    @click.pass_obj
    def create_workspace_share(ctx, workspace, sharee, permission, expire):

        body = {
            "workspace": workspace,
            "sharee": sharee,
            "permission": permission,
        }
        if expire:
            body["expiration"] = expire
        r = ctx["session"].post(
            "workspace/share",
            json=body,
        )
        exit_with(handle_request_error(r))

    cli.add_command(workspace)

    @workspace.command(name="index")
    @click.argument("workspace_id", type=click.STRING)
    @click.option(
        "--minio-mount",
        type=click.Path(dir_okay=True, exists=True),
        help="Path to minio mount on local disk",
    )
    @click.pass_obj
    def index_workspace(ctx, workspace_id, minio_mount):
        # Dynamic, expensive imports
        from workspacesio.common import producers

        ctx = config.getctx(ctx)
        r = ctx.session.get(f"workspace/{workspace_id}")
        if not r.ok:
            exit_with(handle_request_error(r))
        w = schemas.WorkspaceDB(**r.json())
        r = ctx.session.post(f"workspace/{w.id}/crawl")
        if not r.ok:
            exit_with(handle_request_error(r))
        data = indexing_schemas.WorkspaceCrawlRoundResponse(**r.json())
        startfrom = data.crawl_round.last_indexed_key or ""
        root = data.root_credentials.root
        node = data.root_credentials.node
        for batch in producers.minio_buffer_objects(
            producers.minio_recursive_generate_objects(
                node=node,
                root=root,
                workspace=w,
                after=startfrom,
            ),
            buffer_size=100,
        ):
            documents: List[indexing_schemas.IndexDocumentBase] = []
            for obj in batch:
                before = datetime.datetime.utcnow()
                obj.time = "ar"
                doc = producers.minio_transform_object(workspace=w, root=root, obj=obj)
                success, failed = producers.additional_indexes(
                    root=root, workspace=w, doc=doc, node=node
                )
                delta = str(
                    int((datetime.datetime.utcnow() - before).total_seconds() * 1000)
                ).ljust(4)
                click.secho(
                    f"ms={delta} workspace={w.name} analysis={','.join(success)} path={doc.path}",
                    fg="red" if len(failed) else "green",
                )
                documents.append(doc)
            payload = indexing_schemas.IndexBulkAdd(
                documents=documents,
                workspace_id=w.id,
                last_indexed_key=documents[-1].path,
                succeeded=False,
            )
            r = ctx.session.post(
                f"workspace/{w.id}/bulk_index",
                data=payload.json(),
            )
            r.raise_for_status()
        exit_with(
            handle_request_error(
                ctx.session.post(
                    f"workspace/{w.id}/bulk_index",
                    data=indexing_schemas.IndexBulkAdd(
                        documents=[],
                        workspace_id=w.id,
                        succeeded=True,
                    ).json(),
                )
            )
        )



import posixpath
import urllib.parse

import ffmpeg

from . import indexing_schemas, s3utils, schemas


def probe(
    doc: indexing_schemas.IndexDocumentBase,
    node: schemas.StorageNodeOperator,
    root: schemas.WorkspaceRootDB,
    workspace: schemas.WorkspaceDB,
):
    endpoint = node.api_url
    parsed = urllib.parse.urlparse(endpoint)
    host = parsed.netloc
    headerstring = ""
    uri = posixpath.join(
        "/",
        root.bucket,
        s3utils.getWorkspaceKey(workspace, root),
        doc.path.lstrip("/"),
    )
    headers = s3utils.get_s3v4_headers(
        access_key=node.access_key_id,
        secret_key=node.secret_access_key,
        region=node.region_name,
        host=host,
        endpoint=endpoint,
        uri=uri,
    )
    headerstring = "\r\n".join([f"{key}:{val}" for key, val in headers.items()])
    url = urllib.parse.urljoin(endpoint, uri)
    try:
        data = ffmpeg.probe(url, headers=headerstring, show_streams=True)
        if len(data["streams"]):
            streams = data["streams"][0]
            doc.codec_tag_string = streams["codec_tag_string"]
            doc.r_frame_rate = streams["r_frame_rate"]
            doc.width = streams["width"]
            doc.height = streams["height"]
            doc.duration_ts = streams["duration_ts"]
            try:
                print(streams)
                doc.bit_rate = int(streams["bit_rate"])
            except:
                doc.bit_rate = streams["bit_rate"]
        doc.duration_sec = data["format"]["duration"]
        doc.format_name = data["format"]["format_name"]
    except ffmpeg._run.Error as e:
        raise indexing_schemas.ProducerError(e)



"""
Set of schemas that represent POST body of minio/s3 bucket event notifications.
Minio is slightly inconsistent in its specific schema.  These schemas seek
to work for both s3 and minio.

https://docs.aws.amazon.com/AmazonS3/latest/dev/notification-content-structure.html
https://github.com/minio/minio-dotnet/issues/332
"""
import datetime
import uuid
from typing import List, Optional

from pydantic import BaseModel, validator

from . import schemas

INDEX_DOCUMENT_MAPPING = {
    "properties": {
        # Base
        "time": {"type": "date"},
        "size": {"type": "double"},
        "eTag": {"type": "text"},
        "extension": {"type": "keyword"},
        "content_type": {"type": "keyword"},
        "text": {"type": "search_as_you_type"},
        "tag": {"type": "keyword"},
        # Required
        "workspace_id": {
            "type": "keyword",
        },
        "workspace_name": {"type": "keyword"},
        "owner_id": {"type": "keyword"},
        "owner_name": {"type": "keyword"},
        "bucket": {"type": "keyword"},
        "server": {"type": "keyword"},
        "root_path": {"type": "text"},
        "workspace_base_path": {"type": "text"},
        "last_seen_crawl_id": {"type": "keyword"},
        "root_id": {"type": "keyword"},
        "path": {"type": "text"},
        "filename": {"type": "text"},
        "user_shares": {"type": "keyword"},
        # Video
        "codec_tag_string": {"type": "keyword"},
        "width": {"type": "double"},
        "height": {"type": "double"},
        "r_frame_rate": {"type": "keyword"},
        "bit_rate": {"type": "double"},
        "duration_ts": {"type": "double"},
        "duration_sec": {"type": "double"},
        "format_name": {"type": "keyword"},
    }
}


class ProducerError(RuntimeError):
    pass


class IndexDocumentBase(BaseModel):
    """
    The main index document.  Any metadata that can apply to an entire object
    probably belongs in this schema.  It is intended as a union of all single-file
    metadata.  Elastic supports a lot of fields, so this will be fine unless we get into
    the mid hundreds.
    """

    # time of last update.  Note that through s3, creation time is not available
    time: datetime.datetime
    # size in bytes
    size: Optional[int]
    # minio/s3 eTag
    eTag: Optional[str]
    # path is the inner path starting at the workspace boundary
    path: str
    # object name is the file name alone
    filename: str
    # extension is the final extension
    extension: str
    # MIME content type
    content_type: Optional[str]
    # plaintext object contents if utf-8 encodeable
    text: Optional[str]
    # tag unused so far
    tag: Optional[str]

    # Optional Video Metadata
    codec_tag_string: Optional[str]
    width: Optional[int]
    height: Optional[int]
    r_frame_rate: Optional[str]
    bit_rate: Optional[int]
    duration_ts: Optional[int]
    duration_sec: Optional[str]
    format_name: Optional[str]
    # Optional Image Metadata
    # Optional Tabular Metadata
    # Optional Audio Metadata


class IndexDocument(IndexDocumentBase):
    """
    Additional tags that every object gets.  These will be assigned by the server
    if the bulk ingest endpoint is used.
    """

    workspace_id: uuid.UUID
    workspace_name: str
    owner_id: uuid.UUID
    owner_name: str
    bucket: str
    server: str
    root_path: str
    workspace_base_path: str
    last_seen_crawl_id: uuid.UUID
    root_id: uuid.UUID
    user_shares: List[uuid.UUID]


class IndexBase(BaseModel):
    index_type: str
    root_id: uuid.UUID


class IndexCreate(IndexBase):
    pass


class IndexDB(schemas.DBBaseModel, IndexBase):
    root: schemas.WorkspaceRootDB


class IndexBulkAdd(BaseModel):
    """Bulk add records from a workspace into the index"""

    documents: List[IndexDocumentBase]
    workspace_id: uuid.UUID
    last_indexed_key: Optional[str]
    succeeded: Optional[bool]

    @validator("last_indexed_key")
    def validate_last_indexed_key(cls, v, values):
        if len((values["documents"])):
            if v is None:
                raise ValueError("Must specify last_indexed_key when docs are sent")
        return v


class IndexBulkAddedResponse(BaseModel):
    index: IndexDB
    count: int


class EventUserIdentity(BaseModel):
    principalId: str


class Bucket(BaseModel):
    name: str
    ownerIdentity: EventUserIdentity
    arn: str


class EventObject(BaseModel):
    key: str
    size: Optional[int]
    eTag: Optional[str]
    contentType: Optional[str]
    sequencer: str


class S3Event(BaseModel):
    s3SchemaVersion: str
    configurationId: str
    bucket: Bucket
    object: EventObject


class EventNotificationRecord(BaseModel):
    awsRegion: str
    eventName: str
    eventVersion: str
    eventSource: str
    eventTime: datetime.datetime
    userIdentity: EventUserIdentity
    requestParameters: dict
    responseElements: dict
    s3: S3Event


class BucketEventNotification(BaseModel):
    Records: List[EventNotificationRecord]
    EventName: Optional[str]
    Key: Optional[str]


class ElasticUpsertIndexDocument(BaseModel):
    doc: IndexDocument
    doc_as_upsert = True


class WorkspaceCrawlRoundBase(BaseModel):
    workspace_id: uuid.UUID
    start_time: datetime.datetime
    succeeded: bool


class WorkspaceCrawlRoundDB(schemas.DBBaseModel, WorkspaceCrawlRoundBase):
    end_time: Optional[datetime.datetime]
    last_indexed_key: Optional[str]
    total_objects: int
    total_size: int
    workspace: schemas.WorkspaceDB


class WorkspaceCrawlRoundResponse(BaseModel):
    crawl_round: WorkspaceCrawlRoundDB
    root_credentials: schemas.RootCredentials



import json
import os
from typing import Any, Dict, List, Optional, cast

import click
from pydantic import BaseModel
from requests_toolbelt.sessions import BaseUrlSession

from workspacesio.common import schemas


class Config(BaseModel):
    access_key: Optional[str]
    secret_key: Optional[str]
    api_url: str = "http://localhost:8100/api"


class Ctx(BaseModel):
    config: Config
    configPath: str
    session: BaseUrlSession

    class Config:
        arbitrary_types_allowed = True


def make() -> Config:
    return Config()


def getctx(ctx: Dict[str, Any]) -> Ctx:
    return Ctx(**ctx)


def save(ctx: Ctx):
    with open(ctx.configPath, "w") as out:
        out.write(ctx.config.json())


def load_config(path: str) -> Config:
    if os.path.exists(path):
        try:
            with open(path) as config_file:
                return Config(**json.loads(config_file.read()))
        except:
            return make()
    return make()



import os

from setuptools import find_packages, setup

deps = [
    "aiofiles",
    "authlib",
    "boto3",
    "click",
    "click-aliases",
    "colorama",
    "cryptography",
    "databases[postgresql]",
    "elasticsearch>=7.0.0,<8.0.0",
    "fastapi",
    "fastapi-users",
    "fastapi-contrib",
    "ffmpeg-python",
    "gunicorn",
    "jinja2",
    "minio",
    "psycopg2-binary",
    "pydantic",
    "pyjwt",
    "requests",
    "requests-toolbelt",
    "sqlalchemy",
    "starlette",
    "tqdm",
    "uvicorn",
]

setup(
    name="workspacesio",
    version="0.1.0",
    script_name="setup.py",
    python_requires=">3.7",
    zip_safe=False,
    install_requires=deps,
    include_package_data=True,
    packages=find_packages(exclude=["test"]),
    entry_points={
        "console_scripts": [
            "wio=workspacesio.cli:cli",
            "workspaces-create-tables=workspacesio.dev_cli:main",
        ],
    },
)



"""
Helper functions for working with S3 and STS
"""
import base64
import datetime
import hashlib
import hmac
import os
import posixpath
import sys
import urllib.parse
import uuid
from typing import List, Optional, Set, Tuple, TypedDict, Union

import boto3
import minio
from botocore.client import Config

from . import schemas


class Boto3ClientCache:
    """
    There may be many s3 nodes in the cluser.  Once a client has been established,
    cache it for future use.  This class works for sts and s3 type clients.
    """

    def __init__(self):
        self.cache: Dict[str, boto3.Session] = {}

    @staticmethod
    def _get_primary_key(client_type: str, node: schemas.StorageNodeOperator) -> str:
        if not client_type in ["s3", "sts"]:
            raise ValueError(f"{client_type} unsupported by cache")
        primary_key = (
            (
                f"{client_type}{node.region_name}{node.api_url}"
                f"{node.access_key_id}{node.secret_access_key}"
            )
            .lower()
            .encode("utf-8")
        )
        return hashlib.sha256(primary_key).hexdigest()

    def get_client(
        self,
        client_type: str,
        node: schemas.StorageNodeOperator,
    ) -> boto3.Session:
        primary_key_short_sha256 = Boto3ClientCache._get_primary_key(client_type, node)
        client = self.cache.get(primary_key_short_sha256, None)
        if client is None:
            config = Config()
            if client_type == "s3":
                config = Config(signature_version="s3v4")
                client = boto3.client(
                    client_type,
                    region_name=node.region_name,
                    endpoint_url=node.api_url,
                    aws_access_key_id=node.access_key_id,
                    aws_secret_access_key=node.secret_access_key,
                    config=config,
                )
            elif client_type == "sts":
                is_aws = node.assume_role_arn is not None
                api_url = node.api_url
                if node.sts_api_url:
                    api_url = node.sts_api_url
                elif is_aws:
                    api_url = f"https://sts.{node.region_name}.amazonaws.com"
                client = boto3.client(
                    client_type,
                    region_name=node.region_name,
                    endpoint_url=api_url,
                    aws_access_key_id=node.access_key_id,
                    aws_secret_access_key=node.secret_access_key,
                )
            else:
                raise NotImplementedError("Client type not implemented")
            self.cache[primary_key_short_sha256] = client
        return client

    def get_minio_sdk_client(self, node: schemas.StorageNodeOperator) -> minio.Minio:
        primary_key_short_sha256 = (
            Boto3ClientCache._get_primary_key("s3", node) + "minio"
        )
        client = self.cache.get(primary_key_short_sha256, None)
        url = urllib.parse.urlparse(node.api_url)
        if client is None:
            client = minio.Minio(
                url.netloc,
                access_key=node.access_key_id,
                secret_key=node.secret_access_key,
                secure=False,
            )
            self.cache[primary_key_short_sha256] = client
        return client


def sanitize(name: str) -> str:
    # TODO
    return name


def getWorkspaceKey(
    workspace: schemas.WorkspaceDB,
    root: Optional[schemas.WorkspaceRootDB] = None,
) -> str:
    """
    Determine full object prefix for a given workspace
    """
    root_base_path = root.base_path if root else workspace.root.base_path
    inner_path = posixpath.join(workspace.owner.username, sanitize(workspace.name))
    if workspace.base_path != None:
        inner_path = str(workspace.base_path)
    return posixpath.join(
        root_base_path,
        inner_path,
    ).strip("/")


S3V4Headers = TypedDict(
    "S3V4Headers",
    {
        "x-amz-date": str,
        "Authorization": str,
    },
)


def get_s3v4_headers(
    access_key: str,
    secret_key: str,
    region="us-east-1",
    service="s3",
    host="localhost:9000",
    endpoint="http://localhost:9000",
    uri="/",
    request_parameters="",
) -> S3V4Headers:
    method = "GET"
    # Key derivation functions. See:
    # http://docs.aws.amazon.com/general/latest/gr/signature-v4-examples.html#signature-v4-examples-python
    def sign(key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    def getSignatureKey(key, dateStamp, regionName, serviceName):
        kDate = sign(("AWS4" + key).encode("utf-8"), dateStamp)
        kRegion = sign(kDate, regionName)
        kService = sign(kRegion, serviceName)
        kSigning = sign(kService, "aws4_request")
        return kSigning

    # Create a date for headers and the credential string
    t = datetime.datetime.utcnow()
    amzdate = t.strftime("%Y%m%dT%H%M%SZ")
    datestamp = t.strftime("%Y%m%d")  # Date w/o time, used in credential scope

    # ************* TASK 1: CREATE A CANONICAL REQUEST *************
    # http://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html

    # Step 1 is to define the verb (GET, POST, etc.)--already done.

    # Step 2: Create canonical URI--the part of the URI from domain to query
    # string (use '/' if no path)
    canonical_uri = uri

    # Step 3: Create the canonical query string. In this example (a GET request),
    # request parameters are in the query string. Query string values must
    # be URL-encoded (space=%20). The parameters must be sorted by name.
    # For this example, the query string is pre-formatted in the request_parameters variable.
    canonical_querystring = request_parameters

    # Step 4: Create the canonical headers and signed headers. Header names
    # must be trimmed and lowercase, and sorted in code point order from
    # low to high. Note that there is a trailing \n.
    canonical_headers = "host:" + host + "\n" + "x-amz-date:" + amzdate + "\n"

    # Step 5: Create the list of signed headers. This lists the headers
    # in the canonical_headers list, delimited with ";" and in alpha order.
    # Note: The request can include any headers; canonical_headers and
    # signed_headers lists those that you want to be included in the
    # hash of the request. "Host" and "x-amz-date" are always required.
    signed_headers = "host;x-amz-date"

    # Step 6: Create payload hash (hash of the request body content). For GET
    # requests, the payload is an empty string ("").
    payload_hash = hashlib.sha256(("").encode("utf-8")).hexdigest()

    # Step 7: Combine elements to create canonical request
    canonical_request = (
        method
        + "\n"
        + canonical_uri
        + "\n"
        + canonical_querystring
        + "\n"
        + canonical_headers
        + "\n"
        + signed_headers
        + "\n"
        + payload_hash
    )

    # ************* TASK 2: CREATE THE STRING TO SIGN*************
    # Match the algorithm to the hashing algorithm you use, either SHA-1 or
    # SHA-256 (recommended)
    algorithm = "AWS4-HMAC-SHA256"
    credential_scope = datestamp + "/" + region + "/" + service + "/" + "aws4_request"
    string_to_sign = (
        algorithm
        + "\n"
        + amzdate
        + "\n"
        + credential_scope
        + "\n"
        + hashlib.sha256(canonical_request.encode("utf-8")).hexdigest()
    )

    # ************* TASK 3: CALCULATE THE SIGNATURE *************
    # Create the signing key using the function defined above.
    signing_key = getSignatureKey(secret_key, datestamp, region, service)

    # Sign the string_to_sign using the signing_key
    signature = hmac.new(
        signing_key, (string_to_sign).encode("utf-8"), hashlib.sha256
    ).hexdigest()

    # ************* TASK 4: ADD SIGNING INFORMATION TO THE REQUEST *************
    # The signing information can be either in a query string value or in
    # a header named Authorization. This code shows how to use a header.
    # Create authorization header and add to request headers
    authorization_header = (
        algorithm
        + " "
        + "Credential="
        + access_key
        + "/"
        + credential_scope
        + ", "
        + "SignedHeaders="
        + signed_headers
        + ", "
        + "Signature="
        + signature
    )

    # The request can include any headers, but MUST include "host", "x-amz-date",
    # and (for this scenario) "Authorization". "host" and "x-amz-date" must
    # be included in the canonical_headers and signed_headers, as noted
    # earlier. Order here is not significant.
    # Python note: The 'host' header is added automatically by the Python 'requests' library.
    headers: S3V4Headers = {
        "x-amz-date": amzdate,
        "Authorization": authorization_header,
    }
    return headers
